import java.awt.*;
import java.awt.event.*;
import java.awt.image.*;
import javax.swing.*;
import javax.swing.event.*;
import javax.swing.border.*;
import javax.imageio.*;
import java.net.*;
import java.io.*;


/** An instance contains 
  (1) an original image (of class ImageArray),
  (2) a possibly altered (by methods in this instance) version of the original image,
  (3) methods to process the image
  */
public class ImageProcessor {
    
    /** DM provides methods for extracting components of an rgb pixel. */
    public static final DirectColorModel DM= (DirectColorModel) ColorModel.getRGBdefault();
    
    /** The following four constants of this class indicate a color.*/
    /** Color gray */
    public static final int GRAY= 0;
    
    /** Color red */
    public static final int RED= 1;
    
    /** Color green */
    public static final int GREEN= 2;
    
    /** Color blue */
    public static final int BLUE= 3;
    
    private ImageArray originalIm; // The original image, for restoration purposes
    private ImageArray currentIm;  // The altered image
    
    /** Constructor: an instance for im.
      Precondition: im != null. */
    public ImageProcessor(ImageArray im) {
        originalIm= im;
        currentIm= originalIm.copy();
    }
    
    /** = the current image. */
    public ImageArray getCurrentImage() {
        return currentIm;
    }
    
    /** = the original image. */
    public ImageArray getOriginalImage() {
        return originalIm;
    }
    
    /** Invert the current image, replacing each element with its color complement. */
    public void invert() {
        int len= currentIm.getRows() * currentIm.getCols();
        
        // invert all pixels (leave alpha/transparency value alone)
        
        // invariant: pixels 0..p-1 have been complemented.
        for (int p= 0; p < len; p= p+1) {
            int rgb= currentIm.getPixel(p);
            int red= 255 - DM.getRed(rgb);
            int blue= 255 - DM.getBlue(rgb);
            int green= 255 - DM.getGreen(rgb);
            int alpha= DM.getAlpha(rgb);
            currentIm.setPixel(p,
                               (alpha << 24) | (red << 16) | (green << 8) | blue);
        }
    }
    
    /** Transpose the current image.  */
    public void transpose() {
        // Follow this plan: 
        // (1) Create a new ImageArray ia, using currentIM's row-major order array
        //     and rows and columns, but swap the roles of its numbers
        //     of rows and columns.
        // (2) Store the transpose of the currentIm array in ia, using currentIm's
        //     2-parameter getPixel function and ia's 3-parameter setPixel
        //     function.
        // (3) assign ia to currentIm.
        int rows= currentIm.getRows();
        int cols= currentIm.getCols();
        ImageArray ia= new ImageArray(currentIm.getRmoArray(), cols, rows);
        
        //Copy each element ImageArray[r,c] to ia[c,r]
        // invariant: rows 0..r-1 have been copied to ia[.., 0..r-1]
        for (int r= 0; r != rows; r= r+1)
            // invariant: elements [r..0..c-1] have been copied to ia[0..c-1, r]
            for (int c= 0; c != cols; c= c+1) {
            ia.setPixel(c, r, currentIm.getPixel(r,c));
        }
        
        currentIm= ia;
    }
    
    /** Reflect the current image around the horizontal middle. */
    public void hreflect() {
        int rows= currentIm.getRows();
        int cols= currentIm.getCols();
        int h= 0;
        int k= rows-1;
        //invariant: rows 0..h-1 and k+1.. have been inverted
        while (h < k) {
            // Swap row h with row k
            // invariant: pixels 0..c-1 of rows h and k have been swapped
            for (int c= 0; c != cols; c= c+1) {
                currentIm.swapPixels(h, c, k, c);
            }
            
            h= h+1; k= k-1;
        }
    }
    
    /** Reflect the current image around the vertical middle. */
    public void vreflect() {
        int rows= currentIm.getRows();
        int cols= currentIm.getCols();
        int h= 0;
        int k= cols-1;
        //invariant: cols 0..h-1 and k+1.. have been inverted
        while (h < k) {
            // Swap column h with column k
            for (int r= 0; r != rows; r= r+1) {
                currentIm.swapPixels(r, h, r, k);
            }
            
            h= h+1; k= k-1;
        }
    }
    
    /* Filter out color from the current image according to parameter c.<br><br>
     Precondition: c is one of the four color constants GRAY,
     RED, GREEN, BLUE of this class.<br><br>
     
     If c is one of RED, GREEN, BLUE, then filter out that color of the image
     by setting it to 0 in all pixels. If c is GRAY, then filter all color out
     of the image by setting the three color components of each pixel to the
     average of the three.
     
     The alpha component is not changed.
     */
    public void filterOut(int c) {
        if (c != ImageProcessor.GRAY && c != ImageProcessor.RED &&
            c != ImageProcessor.GREEN && c != ImageProcessor.BLUE) {
            System.out.println("invalid color!");
            return;
        }
        
        int len= currentIm.getRows() * currentIm.getCols();
        
        // invariant: Pixels 0..p-1 have been filtered.
        for (int p= 0; p < len; p= p+1) {
            int rgb= currentIm.getPixel(p);
            int red= DM.getRed(rgb);
            int blue= DM.getBlue(rgb);
            int green= DM.getGreen(rgb);
            int alpha= DM.getAlpha(rgb);
            int grey= (int)Math.round((red+green+blue)/3.0);
            int result;
            
            if (c == GRAY)
                result= (alpha << 24) | (grey << 16) | (grey << 8) | grey;
            else if (c == RED)
                result= (alpha << 24) | (0 << 16) | (green << 8) | blue;
            else if (c == GREEN)
                result= (alpha << 24) | (red << 16) | (0 << 8) | blue;
            else
                result= (alpha << 24) | (red << 16) | (green << 8) | 0;
            
            currentIm.setPixel(p, result);          
        }
    }
    
    /** Representation of a message in a picture.
      0. The message is stored as: character '\t', the length of the real message (2 pixels),
      character '\t', and then the characters of the message.
      1. Each character is represented by its 3-digit ascii representation, with leading zeros if necessary.
      2. The length (in characters) of the message, in the range 0..999999, is represented
      in 2 pixels containing len/1000 and len%1000 respectively.
      E.g. if length 543096, one pixel contains 543 and the other 096. 
      3. Let the three digits of an integer to be stored be x*100 + y*10 + z, where each of x, y, z are in
      the range 0..9. then x is stored in red, y in green, and z in blue components. We describe
      how to store x in the red component; the other two are similar.
      
      Change the last digit of the red component to x; if the result is > 255, substract 10.
      */
    
    /** Hide integer n in pixel number p of currentIm.
      Precondition: 0 <= n < 1000 and pixel p exists. */
    private void hide(int n, int p) {
        // Extract the red, green, blue, and alpha components from pixel p of the map.
        int rgb= currentIm.getPixel(p);
        int red= DM.getRed(rgb);
        int green= DM.getGreen(rgb);
        int blue= DM.getBlue(rgb);
        int alpha= DM.getAlpha(rgb);
        
        // Hide n in the red/green/blue components
        red= 10 * (red/10) + n / 100;
        if (red > 255)
            red= red - 10;
        
        green= 10 * (green/10) + (n / 10) % 10;
        if (green > 255)
            green= green - 10;
        
        blue= 10 * (blue/10) + n % 10;
        if (blue > 255)
            blue= blue - 10;
        
        // Store the new pixel in pixel p of the current image
        int result= (alpha << 24) | (red << 16) | (green << 8) | blue;
        currentIm.setPixel(p, result);
    }
    
    /** Hide message m in this image, using the ascii representation of m's chars.
      Return true if this is possible and false if not.
      If m has more than 999999 characters or the picture doesn't have enough
      pixels, return false without storing the message.
      */
    public boolean hide(String m) {
        if (m.length() > 999999 ||
            m.length() + 4 > currentIm.getRows() * currentIm.getCols())
            return false;
        
        hide('\t', 0);
        hide(m.length()/1000, 1);
        hide(m.length()%1000, 2);
        hide('\t', 3);
        int p= 4;
        
        // invariant: The tab char, length of message, tab char, and pixels
        //            0..k-1 are hidden in pixels 0..p-1 of the image.
        for (int k= 0; k < m.length(); k= k+1) {
            hide(m.charAt(k), p);
            p= p+1;            
        }
        return true;
    }
    
    /** = the number n that is hidden in pixel p of the current image. */
    public int getHidden(int p) {
        int rgb= currentIm.getPixel(p);
        int red= DM.getRed(rgb);
        int green= DM.getGreen(rgb);
        int blue= DM.getBlue(rgb);
        return  (red % 10) * 100  +  (green % 10) * 10  +  blue % 10;
    }
    
    /** Extract and return the message hidden in the current image.
      Return null if no message detected. */
    public String reveal() {
        // Extract the first 4 chars, putting the length of the message in len;
        // return false if the first and fourth characters are not tabs
        int t0= getHidden(0);
        int len= getHidden(1)*1000 + getHidden(2);
        int t1= getHidden(3);
        if (t0 != '\t' || t1 != '\t')
            return null;
        
        int p= 4;        
        char[] result= new char[len];
        
        // invariant: All the hidden chars before pixel p have been put
        //            into result[0..k-1]
        for (int k= 0; k < len; k= k+1) {
            result[k]= (char) (getHidden(p));
            p= p+1;
        }
        
        // result contains the message
        return new String(result);
    }
    
    /** Extract and return the message hidden in the current image.
      Return null if no message detected. */
    public String reveal1() {
        // Extract the first 4 chars, putting the length of the message in len;
        // return false if the first and fourth characters are not tabs
        int t0= getHidden(0);
        int len= getHidden(1)*1000 + getHidden(2);
        int t1= getHidden(3);
        if (t0 != '\t' || t1 != '\t')
            return null;
        
        int p= 4;        
        String result= "";
        
        // invariant: All the hidden chars before pixel p have been put
        //            into result[0..k-1]
        for (int k= 0; k < len; k= k+1) {
            result= result + (char) (getHidden(p));
            p= p+1;
        }
        
        // result contains the message
        return result;
    }
    
    /* Assuming the image is broken up into blocks, with nr rows
     and nc columns of blocks, set all pixels of the
     block at position (row, col) to pixel value pixel.
     Precondition: 0  row < nr  and  0  col < nc */
    public void setBlock(int nr, int nc, int row, int col, int pixel) {
        int height= currentIm.getRows() / nr;
        int width= currentIm.getCols() / nc;
        for (int r= 0; r < height; r= r+1) {
            for (int c=0; c < width; c= c+1) {
                currentIm.setPixel(row*height+r, col*width+c, pixel);
            }
        }
    }
    
    /* Assuming the image is broken up into blocks, with nr rows
     of blocks and nc columns of blocks, swap all pixels of
     the blocks at positions (row0, col0) and (row1, col1). */
    public void swapBlocks(int row0, int col0, int row1, int col1, int nr, int nc) {
        int rows= currentIm.getRows();
        int cols= currentIm.getCols();
        int ht= rows/nr;
        int wd= cols/nc;
        for (int r= 0; r < ht; r= r+1) {
            for (int c= 0; c < wd; c= c+1) {
                currentIm.swapPixels(row0*ht+r, col0*wd+c, row1*ht+r, col1*wd+c);
                
            }
        }
    }
    
    /** Restore the original image in the current one */
    public void restore() {
        currentIm= originalIm.copy();
    }
    
    /** Provided file fname does not appear in the current directory, store the
      current image in file fname in the current directory, as a jpg file.
      Write a message on the console indicating whether or not the write was successful. */
    public void writeImage(String fname) throws java.io.IOException {
        File f= new File(fname);
        if (f.exists()) {
            System.out.println("File " + f.getAbsolutePath() + " exists. It was not overwritten.");
            return;
        }
        
        int r= currentIm.getRows();
        int c= currentIm.getCols();
        int roa[]= currentIm.getRmoArray();
        
        // Create an image from roa. The new Container is needed to get an
        // object that has the appropriate function, createImage, in it; that is all.
        Image image= (new Container()).createImage(new MemoryImageSource(c, r, roa, 0, c));
        
        // Obtain a buffered image and the graphics for it. It has to have the right width and height.
        // The third arg of the call below indicates that input into it will be an integer array of
        // rgb colors
        BufferedImage bimage= 
            new BufferedImage(image.getWidth(null), image.getHeight(null), BufferedImage.TYPE_INT_RGB);
        Graphics g= bimage.createGraphics();
        
        
        // put the image into bimage so that it can be written out and dispose of the graphics.
        g.drawImage(image, 0, 0, null);
        g.dispose();
        
        // Finally, write the image onto the file and give the appropriate message
        ImageIO.write(bimage, "jpg", f);
        System.out.println("Image written to " + f.getAbsolutePath());
    }  
    
}
